/* -*- C++ -*-
 *
 * This file is a part of LEMON, a generic C++ optimization library
 *
 * Copyright (C) 2003-2006
 * Egervary Jeno Kombinatorikus Optimalizalasi Kutatocsoport
 * (Egervary Research Group on Combinatorial Optimization, EGRES).
 *
 * Permission to use, modify and distribute this software is granted
 * provided that this copyright notice appears in all copies. For
 * precise terms see the accompanying LICENSE file.
 *
 * This software is provided "AS IS" with no warranty of any kind,
 * express or implied, and with no claim as to its suitability for any
 * purpose.
 *
 */

#ifndef LEMON_BITS_BASE_EXTENDER_H
#define LEMON_BITS_BASE_EXTENDER_H

#include <lemon/bits/invalid.h>
#include <lemon/error.h>

#include <lemon/bits/map_extender.h>
#include <lemon/bits/default_map.h>

#include <lemon/concept_check.h>
#include <lemon/concept/maps.h>

///\ingroup graphbits
///\file
///\brief Extenders for the graph types
namespace lemon {

  /// \ingroup graphbits
  ///
  /// \brief BaseExtender for the UGraphs
  template <typename Base>
  class UndirGraphExtender : public Base {

  public:

    typedef Base Parent;
    typedef typename Parent::Edge UEdge;
    typedef typename Parent::Node Node;

    typedef True UndirectedTag;

    class Edge : public UEdge {
      friend class UndirGraphExtender;

    protected:
      bool forward;

      Edge(const UEdge &ue, bool _forward) :
        UEdge(ue), forward(_forward) {}

    public:
      Edge() {}

      /// Invalid edge constructor
      Edge(Invalid i) : UEdge(i), forward(true) {}

      bool operator==(const Edge &that) const {
	return forward==that.forward && UEdge(*this)==UEdge(that);
      }
      bool operator!=(const Edge &that) const {
	return forward!=that.forward || UEdge(*this)!=UEdge(that);
      }
      bool operator<(const Edge &that) const {
	return forward<that.forward ||
	  (!(that.forward<forward) && UEdge(*this)<UEdge(that));
      }
    };



    using Parent::source;

    /// Source of the given Edge.
    Node source(const Edge &e) const {
      return e.forward ? Parent::source(e) : Parent::target(e);
    }

    using Parent::target;

    /// Target of the given Edge.
    Node target(const Edge &e) const {
      return e.forward ? Parent::target(e) : Parent::source(e);
    }

    /// \brief Directed edge from an undirected edge.
    ///
    /// Returns a directed edge corresponding to the specified UEdge.
    /// If the given bool is true the given undirected edge and the
    /// returned edge have the same source node.
    static Edge direct(const UEdge &ue, bool d) {
      return Edge(ue, d);
    }

    /// Returns whether the given directed edge is same orientation as the
    /// corresponding undirected edge.
    ///
    /// \todo reference to the corresponding point of the undirected graph
    /// concept. "What does the direction of an undirected edge mean?"
    static bool direction(const Edge &e) { return e.forward; }


    using Parent::first;
    using Parent::next;

    void first(Edge &e) const {
      Parent::first(e);
      e.forward=true;
    }

    void next(Edge &e) const {
      if( e.forward ) {
	e.forward = false;
      }
      else {
	Parent::next(e);
	e.forward = true;
      }
    }

    void firstOut(Edge &e, const Node &n) const {
      Parent::firstIn(e,n);
      if( UEdge(e) != INVALID ) {
	e.forward = false;
      }
      else {
	Parent::firstOut(e,n);
	e.forward = true;
      }
    }
    void nextOut(Edge &e) const {
      if( ! e.forward ) {
	Node n = Parent::target(e);
	Parent::nextIn(e);
	if( UEdge(e) == INVALID ) {
	  Parent::firstOut(e, n);
	  e.forward = true;
	}
      }
      else {
	Parent::nextOut(e);
      }
    }

    void firstIn(Edge &e, const Node &n) const {
      Parent::firstOut(e,n);
      if( UEdge(e) != INVALID ) {
	e.forward = false;
      }
      else {
	Parent::firstIn(e,n);
	e.forward = true;
      }
    }
    void nextIn(Edge &e) const {
      if( ! e.forward ) {
	Node n = Parent::source(e);
	Parent::nextOut(e);
	if( UEdge(e) == INVALID ) {
	  Parent::firstIn(e, n);
	  e.forward = true;
	}
      }
      else {
	Parent::nextIn(e);
      }
    }

    void firstInc(UEdge &e, bool &d, const Node &n) const {
      d = true;
      Parent::firstOut(e, n);
      if (e != INVALID) return;
      d = false;
      Parent::firstIn(e, n);
    }

    void nextInc(UEdge &e, bool &d) const {
      if (d) {
	Node s = Parent::source(e);
	Parent::nextOut(e);
	if (e != INVALID) return;
	d = false;
	Parent::firstIn(e, s);
      } else {
	Parent::nextIn(e);
      }
    }

    Node nodeFromId(int id) const {
      return Parent::nodeFromId(id);
    }

    Edge edgeFromId(int id) const {
      return direct(Parent::edgeFromId(id >> 1), bool(id & 1));
    }

    UEdge uEdgeFromId(int id) const {
      return Parent::edgeFromId(id >> 1);
    }

    int id(const Node &n) const {
      return Parent::id(n);
    }

    int id(const UEdge &e) const {
      return Parent::id(e);
    }

    int id(const Edge &e) const {
      return 2 * Parent::id(e) + int(e.forward);
    }

    int maxNodeId() const {
      return Parent::maxNodeId();
    }

    int maxEdgeId() const {
      return 2 * Parent::maxEdgeId() + 1;
    }

    int maxUEdgeId() const {
      return Parent::maxEdgeId();
    }


    int edgeNum() const {
      return 2 * Parent::edgeNum();
    }

    int uEdgeNum() const {
      return Parent::edgeNum();
    }

    Edge findEdge(Node source, Node target, Edge prev) const {
      if (prev == INVALID) {
	UEdge edge = Parent::findEdge(source, target);
	if (edge != INVALID) return direct(edge, true);
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return direct(edge, false);
      } else if (direction(prev)) {
	UEdge edge = Parent::findEdge(source, target, prev);
	if (edge != INVALID) return direct(edge, true);
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return direct(edge, false);	
      } else {
	UEdge edge = Parent::findEdge(target, source, prev);
	if (edge != INVALID) return direct(edge, false);	      
      }
      return INVALID;
    }

    UEdge findUEdge(Node source, Node target, UEdge prev) const {
      if (prev == INVALID) {
	UEdge edge = Parent::findEdge(source, target);
	if (edge != INVALID) return edge;
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return edge;
      } else if (Parent::source(prev) == source) {
	UEdge edge = Parent::findEdge(source, target, prev);
	if (edge != INVALID) return edge;
	edge = Parent::findEdge(target, source);
	if (edge != INVALID) return edge;	
      } else {
	UEdge edge = Parent::findEdge(target, source, prev);
	if (edge != INVALID) return edge;	      
      }
      return INVALID;
    }
  };

}

#endif
